--- old/usr/src/lib/libzfs/common/libzfs_sendrecv.c	2013-06-17 09:37:34.342651478 +0000
+++ new/usr/src/lib/libzfs/common/libzfs_sendrecv.c	2013-06-17 09:37:34.348625391 +0000
@@ -723,7 +723,7 @@
 	sd->parent_fromsnap_guid = 0;
 	VERIFY(0 == nvlist_alloc(&sd->parent_snaps, NV_UNIQUE_NAME, 0));
 	VERIFY(0 == nvlist_alloc(&sd->snapprops, NV_UNIQUE_NAME, 0));
-	(void) zfs_iter_snapshots(zhp, send_iterate_snap, sd);
+	(void) zfs_iter_snapshots_sorted(zhp, send_iterate_snap, sd);
 	VERIFY(0 == nvlist_add_nvlist(nvfs, "snaps", sd->parent_snaps));
 	VERIFY(0 == nvlist_add_nvlist(nvfs, "snapprops", sd->snapprops));
 	nvlist_free(sd->parent_snaps);
@@ -1936,11 +1936,11 @@
     recvflags_t *flags, nvlist_t *stream_nv, avl_tree_t *stream_avl,
     nvlist_t *renamed)
 {
-	nvlist_t *local_nv;
+	nvlist_t *local_nv, *deleted = NULL;
 	avl_tree_t *local_avl;
 	nvpair_t *fselem, *nextfselem;
 	char *fromsnap;
-	char newname[ZFS_MAXNAMELEN];
+	char newname[ZFS_MAXNAMELEN], guidname[32];
 	int error;
 	boolean_t needagain, progress, recursive;
 	char *s1, *s2;
@@ -1960,6 +1960,8 @@
 	    recursive, &local_nv, &local_avl)) != 0)
 		return (error);
 
+	deleted = fnvlist_alloc();
+
 	/*
 	 * Process deletes and renames
 	 */
@@ -2031,6 +2033,7 @@
 			case -1:
 				fsavl_destroy(local_avl);
 				nvlist_free(local_nv);
+				nvlist_free(deleted);
 				return (-1);
 			}
 			/*
@@ -2070,6 +2073,9 @@
 					needagain = B_TRUE;
 				else
 					progress = B_TRUE;
+				(void) sprintf(guidname, "%llu",
+				    (u_longlong_t)thisguid);
+				fnvlist_add_boolean(deleted, guidname);
 				continue;
 			}
 
@@ -2125,6 +2131,9 @@
 				needagain = B_TRUE;
 			else
 				progress = B_TRUE;
+			(void) sprintf(guidname, "%llu",
+			    (u_longlong_t)parent_fromsnap_guid);
+			fnvlist_add_boolean(deleted, guidname);
 			continue;
 		}
 
@@ -2147,6 +2156,24 @@
 		s2 = strrchr(stream_fsname, '/');
 
 		/*
+		 * Check if we're going to rename based on parent guid change
+		 * and the current parent guid was also deleted. If it was then
+		 * the rename will fail so avoid this and force an early retry
+		 * to determine the new parent_fromsnap_guid.
+		 */
+		if (stream_parent_fromsnap_guid != 0 &&
+		    parent_fromsnap_guid != 0 &&
+		    stream_parent_fromsnap_guid != parent_fromsnap_guid) {
+			(void) sprintf(guidname, "%llu",
+			    (u_longlong_t)parent_fromsnap_guid);
+			if (nvlist_exists(deleted, guidname)) {
+				progress = B_TRUE;
+				needagain = B_TRUE;
+				goto doagain;
+			}
+		}
+
+		/*
 		 * Check for rename. If the exact receive path is specified, it
 		 * does not count as a rename, but we still need to check the
 		 * datasets beneath it.
@@ -2200,8 +2227,10 @@
 		}
 	}
 
+doagain:
 	fsavl_destroy(local_avl);
 	nvlist_free(local_nv);
+	nvlist_free(deleted);
 
 	if (needagain && progress) {
 		/* do another pass to fix up temporary names */
